#include "AESLib.h" //AES Encryption library
//Make sure to copy all the library files in the same folder of this arduino sketch

#include <EEPROM.h> //Arduino library to ease the internal EEPROM usage
//Arduino Nano ATmega328P chip has 1024 Bytes of EEPROM memory

//Macros needed for bit value changing of each password status byte
#define sbi(Byte,bit) (Byte |= (1<<bit)) //set bit in port
#define cbi(Byte,bit) (Byte &= ~(1<<bit)) //clear bit in port

bool rbit(uint8_t Byte, uint8_t bit){ //Reads a bit value inside a byte
  return Byte & (1 << bit);
}
void EEPROMSetBit(int Address, uint8_t bit){
  byte EEPROMModifiedByte = EEPROM.read(Address);
  sbi(EEPROMModifiedByte, bit);
  EEPROM.write(Address, EEPROMModifiedByte); 
}
void EEPROMResetBit(int Address, uint8_t bit){
  byte EEPROMModifiedByte = EEPROM.read(Address);
  cbi(EEPROMModifiedByte, bit);
  EEPROM.write(Address, EEPROMModifiedByte); 
}

uint8_t key[] = {'D','U','M','M','Y','T','E','S','T','K','E','Y','1','2','3','!'}; //Change with your 16 byte key here
char devicePassword[] = "1234567890"; //change this password for your vault

bool unlocked = 0; //if 0 then arduino has is locked, turns 1 after entering the correct vault password
String enteredData; //stores user entered string texts 
char enteredDataArray[50]; //buffer to store entered string from user for further use 
char outputData[50]; //stores data that gets read from the EEPROM
char EEPROMString[50]; //char array that gets stored in the EEPROM
char encryptionBuffer[50]; //char array buffer to deal with aes128_enc_single() and aes128_dec_single() functions
uint8_t current_screen = 0; //if the vault is unlocked, this variable tells which user screen is printed on the serial
//user screens as (1 - request password, 2 - add password, 3 - change password, 4 - delete paswword, 5 - list websites names)

int passwordAddress[10] = {0,100, 200, 300, 400, 500, 600, 700, 800, 900};
/* This vault can save 10 passwords
 *  and the following array contains each password address in the chip EEPROM
 *  the 100 bytes saved for each password are distributed as following below
 *  [-----20 bytes for website/password name----][----30 bytes for login data as email address/username----][----40 bytes for the encrypted password----][-1 byte for password status-][--9 bytes for further updates if needed--]
 *  password name bytes start from passwordAddress[i*10] to passwordAddress[i*10+19] example: from 0 to 19 or 200 to 219 (20 bytes)
 *  password login data start from passwordAddress[i*10+20] to passwordAddress[i*10+49] exapmle from 20 to 49 or from 220 to 249 (30 bytes)
 *  encrypted password  start from passwordAddress[i*10+50] to passwordAddress[i*10+89] exapmle from 50 to 89 or from 250 to 289 (40 bytes)
 *  password status byte is at passwordAddress[i*10+90] example address 90 or address 290 (1 byte)
 *  Status byte [0b 1111 111A], A: Password location availability [1 if empty and 0 if occupied by another password]
 *  Maximum website name length should be less than 19 characters
 *  Maximum website login length should be less than 29 characters
 *  Maximum password length should be less than 49 characters
 */

void waitForPassword(); //Checks if the user entered the coorect vault password or not
bool scanForInput(); //A function that reads user input data and decides if its a command or a text, then changes user screen
void mainScreen(); //Default empty screen
void passwordRequest(); //Screen for entering password name
void passwordAdd(); //Screen for registering new password
void passwordChange(); //Screen for changing password value and keeping the same password name and login
void passwordDelete(); //Screen for deleting an existing password
void passwordList(); //Screen for listing all of the saved passwords names
void EEPROMWriteString(int Address,char* stringIn); //write String to specific EEPROM address
void EEPROMReadString(int Address,char* stringOut); //read a string from a specific EEPROM address until reaching the termination character (0x00)
  
void setup() {
Serial.begin(115200);
}

void loop() {
while (unlocked){ //all the unlocked vault operations happen inside the while loop
  scanForInput();
  
  switch(current_screen){
    case 0: mainScreen(); 
        break;
    case 1: passwordRequest();
        break;
    case 2: passwordAdd();
        break;
    case 3: passwordChange();
        break;
    case 4: passwordDelete();
        break;
    case 5: passwordList();
        break;
  }
}
  waitForPassword(); //if vault locked then wait for password entry
}

void waitForPassword(){
  current_screen = 0;
  unlocked = 0;
  if (Serial.available() > 0){
    enteredData = Serial.readStringUntil('\n');
    if (enteredData == devicePassword){
      unlocked = 1; //unlock the vault when correct password is entered
      current_screen = 0;
      enteredData = ""; // reset the input scan string
      Serial.println("Correct Password");
    }
  }
  enteredData = ""; // reset the input scan string
}

bool scanForInput(){
  if (Serial.available() > 0){
    enteredData = Serial.readStringUntil('\n');
    for (int i = 0; i<50; i++){ enteredDataArray[i] = 0x00; } //clearing entered data char array and saving entere data string in it
    for (int i = 0; i<50; i++){
    enteredDataArray[i] = enteredData[i];
    }
    if ( enteredData[1] == 0 ){     /*checks if the entered data is a user command or a user text, 
                                    *if it is a user single letter command then the second character of the string 
                                    *will be equal to 0
                                    */
      switch(enteredData[0]){ //displays the user screen corresponding to the command input
        default : enteredData = ""; waitForPassword(); break; //if unrecognized character is entered then return back to the vault password input screen
        case 'F': enteredData = ""; waitForPassword(); break; //return back to the vault password input screen
        case 'M': enteredData = ""; mainScreen(); break; //go to the default Main screen
        case 'R': enteredData = ""; passwordRequest(); break; //go to the password name Entering screen
        case 'A': enteredData = ""; passwordAdd(); break; //go to the password Add screen
        case 'C': enteredData = ""; passwordChange(); break; //go to the password Change screen
        case 'D': enteredData = ""; passwordDelete(); break; //go to the password Delete screen
        case 'L': enteredData = ""; passwordList(); break; //go to the password name Listing screen
      }
      return 1;
    } 
  }else {
    return 0;
  }
}

void mainScreen(){
  bool scanInput = 0;
  Serial.println("Main Screen");
  while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
}

void passwordRequest(){
  bool scanInput = 0;
  Serial.println("Enter password name:");
  while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
  Serial.println(enteredData);
  for (int i = 0; i<=900; i+=100){
    for (int i = 0; i<50; i++){outputData[i] = 0x00;} //clear outputData array  
    EEPROMReadString(i, outputData);;
    if(strcmp(outputData, enteredDataArray) == 0){ //if both entered string and string read out from the eeprom are both equal then the output of the strcmp function is 0
      Serial.println("Password found");
      Serial.print("PW Name: "); Serial.println(enteredDataArray);
      Serial.print("PW Login: "); EEPROMReadString(i+20, outputData); Serial.println(outputData);
      Serial.print("Password: "); EEPROMReadString(i+50, outputData); //reading encrypted password from EEPROM
                                  aes128_dec_single(key, outputData); //decryption step
                                  Serial.println(outputData);
      break;
    }
  }
  enteredData = "";
  current_screen = 0;
}

void passwordAdd(){
  bool scanInput = 0;
  int EEPROMLocation = 0;
  for (int i = 0; i <= 900; i+=100){
    if (rbit(EEPROM.read(i+90), 0) == 1){ //checks for an empty password place to store the new password
      EEPROMLocation = i;
      Serial.println("Register password name:"); //User inputs the website that the password is for (Don't enter more than 20 characters)
      while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
        enteredData.toCharArray(EEPROMString, 50);
        EEPROMWriteString(i, EEPROMString);
        for (int i = 0; i < 50; i++){ //Resetting EEPROMString
          EEPROMString[i] = 0xFF;
        }
        scanInput = 0;
        enteredData = "";
      Serial.println("Register password login:"); //User enters login email address or username of the password (Don't enter more than 30 characters)
      while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
        enteredData.toCharArray(EEPROMString, 50);
        EEPROMWriteString(i+20, EEPROMString);
        for (int i = 0; i < 50; i++){
          EEPROMString[i] = 0xFF;
        }
        scanInput = 0;
        enteredData = "";
      Serial.println("Register password:"); //User inputs the actual password (Don't enter more than 50 characters)
      while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
        enteredData.toCharArray(EEPROMString, 50);
        sprintf(encryptionBuffer, "%-16s", EEPROMString); //right padding using 16 spaces character because encryptio and ecryption needs 16 characaters long password at least
        aes128_enc_single(key, encryptionBuffer); //encryption step
        EEPROMWriteString(i+50, encryptionBuffer); //Save encrypted password to EEPROM
        for (int i = 0; i < 50; i++){
          EEPROMString[i] = 0xFF;
        }
        scanInput = 0;
        enteredData = "";
        EEPROMResetBit(i+90, 0); //Reset 0 bit in the status byte to mark that this place contains a password and not empty
      break; 
    } else if (i+90 == 990) { //if it reached byte address 990 without finding any empty space, the user needs to delete a password to empty space
      Serial.println("Vault full, Empty space");
      current_screen = 0;
    }
  }
}

void passwordChange(){
  bool scanInput = 0;
  Serial.println("Enter password name:");
  while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
  for (int i = 0; i<=900; i+=100){
    for (int i = 0; i<50; i++){outputData[i] = 0x00;} //clear outputData array  
    EEPROMReadString(i, outputData);
    if(strcmp(outputData, enteredDataArray) == 0){ //if both entered string and string read out from the eeprom are both equal then the output of the strcmp function is 0
      Serial.println("Password found \nEnter new password:");
        scanInput = 0;
        enteredData = "";
        while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input for new password
         for (int j = i+50; j < i+90; j++){ //erases only the password value section
         EEPROM.write(j, 0xFF);
         } 
        enteredData.toCharArray(EEPROMString, 50);
        sprintf(encryptionBuffer, "%-16s", EEPROMString); //right padding using 16 spaces character because encryptio and ecryption needs 16 characaters long password at least
        aes128_enc_single(key, encryptionBuffer); //encryption step
        EEPROMWriteString(i+50, encryptionBuffer); //Save encrypted password to EEPROM
        Serial.println("Password Changed");
      break;
    }
  }
  scanInput = 0;
  enteredData = "";
  current_screen = 0;
}

void passwordDelete(){
  bool scanInput = 0;
  Serial.println("Enter password name:");
  while (scanInput == 0){ scanInput = scanForInput(); } //wait for user input
  for (int i = 0; i<=900; i+=100){
    for (int i = 0; i<50; i++){outputData[i] = 0x00;} //clear outputData array  
    EEPROMReadString(i, outputData);
    if(strcmp(outputData, enteredDataArray) == 0){ //if both entered string and string read out from the eeprom are both equal then the output of the strcmp function is 0
      Serial.println("Password found");
      for (int j = i; j < i+100; j++){ //erases the complete password section encluding password name, login data and password value
        EEPROM.write(j, 0xFF);
      }
      EEPROMSetBit(i+90, 0); //Set 0 bit in the status byte to mark that this place is empty
      Serial.println("Password deleted");
      break;
    }
  }
  scanInput = 0;
  enteredData = "";
  current_screen = 0;
}

void passwordList(){
  Serial.println("Saved password names:");
  for (int i = 0; i<=900; i+=100){
    Serial.print((i/100)+1); Serial.print("-"); 
    for (int i = 0; i<50; i++){outputData[i] = 0x00;} //clear outputData array  
    EEPROMReadString(i, outputData);
    Serial.println(outputData);
  }
  current_screen = 0;
}

void EEPROMWriteString(int Address,char* stringIn){ //write String to specific EEPROM address
  for(int i = 0; i < strlen(stringIn); i++){
      EEPROM.write(i+Address, stringIn[i]); 
  }
  EEPROM.write(Address+strlen(stringIn), 0x00); //add termination character
}

void EEPROMReadString(int Address,char* stringOut){ //read a string from a specific EEPROM address until reaching the termination character (0x00)
  int EEPROMByte = 0x01;
  int k = 0;
  while ((EEPROMByte != 0x00) && (EEPROMByte != 0XFF)){
    EEPROMByte = EEPROM.read(Address+k);
    stringOut[k] = EEPROMByte;
    k++;
  }
}


